﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text.RegularExpressions;
using System.Diagnostics;
using System.Collections;
using System.Net.NetworkInformation;

namespace cloudtreeReference
{
    public class NetTcpServer
    {
        public delegate void PacketUpdateDelegate(ref byte[] data);
        public event PacketUpdateDelegate OnUpdate;

        private Socket tcpServer = null;
        private Socket tcpClient = null;
        private Socket tcpClient_Dummy = null;
        private IPEndPoint tcpServerEP = null;

        private ArrayList listeningSocket = new ArrayList();
        private string rcvIP;
        private int rcvPort = -1, sendPort = -1;
        
        private byte[] data;
        private int maxLength;

        private Thread TcpAcceptThread = null;
        private Thread TcpRecvThread = null;

        private bool m_bTcpThreadAllive = false;
        private bool m_bTcpAcceptThreadAlive = false;
        private bool bExitClientThread = false;
        private bool bExitAcceptThread = false;

        private int _lastSocketErrorNum = 0;
        private string _lasSocketErrorMessage;

        public const int FIONREAD = 0x4004667F;
        public const int DEFAULT_RECV_TIMEOUT = 3000;
        public const int RESULT_RECV_TIMEOUT = 60 * 1000; // 1분 대기 

        public NetTcpServer()
        {
            rcvIP = GetRealIpAddress().ToString();
        }

        public void SetRcvPort(int port)
        {
            rcvPort = port;
        }

        public NetTcpServer(string _ip, int _port)
        {
            rcvIP = _ip;
            rcvPort = _port;
        }

        public void SetDataMaxLength(int length)
        {
            maxLength = length;
            data = new byte[length];
        }

        public bool Connect()
        {
            if (tcpServer != null)
            {
                tcpServer.Close();
                tcpServer = null;
            }
            if (rcvPort == -1)
                return false;
            else
            {
                tcpServer = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                tcpServerEP = new IPEndPoint(IPAddress.Parse(rcvIP), rcvPort);

                try
                {
                    tcpServer.Bind(tcpServerEP);
                    tcpServer.Listen(10);
                }
                catch (System.Exception ex) { }

                listeningSocket.Add(tcpServer); // 필요한가? 추후 검토.
            }

            bExitAcceptThread = false;
            bExitClientThread = false;

            TcpAcceptThread = new Thread(new ThreadStart(AcceptThread));
            TcpAcceptThread.Start();
            //TcpRecvTimeout(5000);

            return tcpServer.IsBound;
        }

        // TCP Client 접속 대기 
        public void AcceptThread()
        {
            m_bTcpAcceptThreadAlive = true;
            while (!bExitAcceptThread)
            {
                try
                {
                    tcpClient = tcpServer.Accept();
                    if (tcpClient.Connected)
                    {
                        bExitClientThread = false;
                        TcpRecvThread = new Thread(new ParameterizedThreadStart(TcpRecvFunc));
                        TcpRecvThread.Start(tcpClient);
                        //tcpClient.ReceiveTimeout = 5000; // 5초간 메세지 전송 없으면 리시브 스레드 자동 종료
                    }

                    if (bExitAcceptThread)
                        break;
                }
                catch (Exception e)
                {
                    Trace.WriteLine(e.ToString());
                }
            }
            m_bTcpAcceptThreadAlive = false;
            Trace.WriteLine("Accept Thread Terminated.");
        }

        // TCP recv 대기 시간: ms
        public void TcpRecvTimeout(int ms)
        {
            if (tcpServer == null)
                return;

            try
            {
                tcpServer.ReceiveTimeout = ms;
            }
            catch (SocketException se)
            {
                Trace.WriteLine("Exception: TcpRecvTimeout: " + se.Message);
            }
        }

        public void TcpRecvFunc(object client)
        {
            Socket sock = client as Socket;
            m_bTcpThreadAllive = true;
            int len = 0;

            while (!bExitClientThread)
            {
                try
                {
                    len = sock.Receive(data, 0, maxLength, SocketFlags.None);
                    
                    if (OnUpdate != null && len > 0)
                        OnUpdate(ref data);

                    TcpClearPendingBuffer();
                    System.Array.Clear(data, 0, maxLength);
                }
                catch (System.Exception ex)
                {
                    bExitClientThread = true;
                    break;
                }
            }
            m_bTcpThreadAllive = false;
        }

        public bool GetConnStatus()
        {
            return tcpServer.IsBound;
        }

        public int SendData(byte[] sendData, int length)
        {
            int len = 0;

            try
            {
                if (tcpClient == null)
                    return -1;
                else if (!tcpClient.Connected)
                    return -1;

                TcpClearPendingBuffer(); // 소켓 버퍼 비우기

                len = tcpClient.Send(sendData, 0, length, SocketFlags.None);
            }
            catch (SocketException se)
            {
                this.setLastErrorMsg(se);
            }

            return len;
        }

        private object _lockEvent = new object();   // lock object
        public bool Disconnect()
        {
            bool rtn = false;

            bExitAcceptThread = true;

            tcpClient_Dummy = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPAddress serverIp = GetRealIpAddress();
            IPEndPoint serverEndPoint = new IPEndPoint(serverIp, rcvPort);
            tcpClient_Dummy.Connect(serverEndPoint);
            tcpClient_Dummy.Shutdown(SocketShutdown.Both);
            tcpClient_Dummy.Close();
            tcpClient_Dummy = null;

            if (tcpServer == null)
                return true;

            lock (_lockEvent)
            {
                if (tcpClient != null)
                {
                    if (tcpClient.Connected)
                    {
                        try
                        {
                            tcpClient.Shutdown(SocketShutdown.Both);
                            tcpClient.Close();
                            tcpClient.Dispose();
                            tcpClient = null;
                        }
                        catch (System.Exception ex) { }
                    }
                }


                if (tcpServer.IsBound)
                {
                    try
                    {
                        bExitAcceptThread = true;
                        tcpServer.Shutdown(SocketShutdown.Both);

                        // 2015.03.13 수정
                        // Disconnect 함수는 수신측에서 응답 있어야 리턴되어 hang 발생될 수 있어서 수정 
                        // Receive 호출중 Disconnect 호출시 Receive 응답후 Disconnect 리턴됨
                        //sockTcp.Disconnect(true); 
                    }
                    catch (SocketException se)
                    {
                        //Trace.WriteLine("Exception: Disconnect: " + exp.Message);
                        setLastErrorMsg(se);
                    }

                    tcpServer.Close();
                    tcpServer = null;
                    rtn = true;
                }
            }

            return rtn;
        }

        private void TcpClearPendingBuffer()
        {
            byte[] outValue = BitConverter.GetBytes(0);

            if (tcpServer == null)
                return;

            try
            {
                tcpServer.IOControl(FIONREAD, null, outValue);

                int pendingSize = BitConverter.ToInt32(outValue, 0);

                byte[] recvBuffer = new byte[maxLength];
                int len = 0;
                for (int i = 0; i < pendingSize; i++)
                {
                    if ((len = recvPacket(ref recvBuffer, 1)) <= 0)
                        return;
                }
            }
            catch (SocketException se)
            {
                this.setLastErrorMsg(se);
            }
        }

        private int recvPacket(ref byte[] buffer, int length)
        {
            int len = 0;

            try
            {
                if (tcpClient == null)
                    return -1;
                else if (!tcpClient.Connected)
                    return -1;

                //Trace.WriteLine("recvPacket 1: " + socketClient.Connected.ToString());
                len = tcpClient.Receive(buffer, 0, length, SocketFlags.None);
                //Trace.WriteLine("recvPacket 2: " + socketClient.Connected.ToString());
            }
            catch (SocketException se)
            {
                Trace.WriteLine("recvPacket Exception: " + tcpClient.Connected.ToString());
                this.setLastErrorMsg(se);
            }

            return len;
        }

        // 소켓 에러 저장
        private void setLastErrorMsg(SocketException se)// string msg)
        {
            _lastSocketErrorNum = se.ErrorCode;
            _lasSocketErrorMessage = se.Message;// msg;

            Trace.WriteLine("setLastError(" + se.ErrorCode.ToString() + "): " + se.Message);
        }

        #region GetRealIpAddress
        private IPAddress GetRealIpAddress()
        {
            IPAddress gateway = FindGetGatewayAddress();

            if (gateway == null)
                return null;

            IPAddress[] pIPAddress = Dns.GetHostAddresses(Dns.GetHostName());

            foreach (IPAddress address in pIPAddress)
                if (IsAddressOfGateway(address, gateway))
                    return address;

            return null;
        }

        private bool IsAddressOfGateway(IPAddress address, IPAddress gateway)
        {
            if (address != null && gateway != null)
                return IsAddressOfGateway(address.GetAddressBytes(), gateway.GetAddressBytes());

            return false;
        }

        private bool IsAddressOfGateway(byte[] address, byte[] gateway)
        {
            if (address != null && gateway != null)
            {
                int gwLen = gateway.Length;
                if (gwLen > 0)
                {
                    if (address.Length == gateway.Length)
                    {
                        --gwLen;
                        int counter = 0;
                        for (int i = 0; i < gwLen; i++)
                        {
                            if (address[i] == gateway[i])
                                ++counter;
                        }
                        return (counter == gwLen);
                    }
                }
            }
            return false;
        }

        private IPAddress FindGetGatewayAddress()
        {
            IPGlobalProperties ipGlobProps = IPGlobalProperties.GetIPGlobalProperties();

            foreach (NetworkInterface ni in NetworkInterface.GetAllNetworkInterfaces())
            {
                IPInterfaceProperties ipInfProps = ni.GetIPProperties();
                foreach (GatewayIPAddressInformation gi in ipInfProps.GatewayAddresses)
                    return gi.Address;
            }
            return null;
        }
        #endregion

    }
}
