﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text.RegularExpressions;
using System.Diagnostics;
using System.Net.NetworkInformation;

namespace cloudtreeReference
{
    public class NetUdpMulticast
    {
        public delegate void PacketUpdateDelegate(ref byte[] data);
        public event PacketUpdateDelegate OnUpdate;

        private UdpClient udpMulticast = null;
        private IPEndPoint udpBindEP = null;
        private IPEndPoint udpRcvEP = null;
        private IPEndPoint sendEP = null;

        private string multicastIP;
        private int multicastPort = -1;
        private bool m_bUdpThreadAllive = false;
        private bool m_bUdpTerminate = false;
        private byte[] data;

        private Thread UdpThread = null;

        public NetUdpMulticast()
        {
            multicastIP = GetRealIpAddress().ToString();
        }

        public void SetRcvPort(int port)
        {
            multicastPort = port;
        }

        public NetUdpMulticast(string _IP, int _Port)
        {
            multicastIP = _IP;
            multicastPort = _Port;
        }

        //public String getIP()
        //{
        //    Regex regex = new Regex(@"^(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])$");
        //    foreach (System.Net.IPAddress ip in System.Net.Dns.GetHostEntry(System.Net.Dns.GetHostName()).AddressList)
        //    {
        //        if (regex.IsMatch(ip.ToString()))
        //        {
        //            return ip.ToString();
        //        }
        //    }
        //    return null;
        //}

        public void SetDataMaxLength(int length)
        {
            data = new byte[length];
        }

        public bool Connect()
        {
            if (udpMulticast != null)
            {
                udpMulticast.Close();
                udpMulticast = null;
            }


            if (multicastPort == -1)
                return false;
            else
            {
                udpMulticast = new UdpClient();
                // 멀티캐스트 send용 IPEndPoint
                sendEP = new IPEndPoint(IPAddress.Parse(multicastIP), multicastPort);

                // 멀티캐스트 그룹 Bind
                udpBindEP = new IPEndPoint(IPAddress.Any, multicastPort);
                udpMulticast.Client.Bind(udpBindEP);

                // 멀티캐스트 rcv용 IPEndPoint
                IPAddress _IP = IPAddress.Parse(multicastIP);
                udpMulticast.JoinMulticastGroup(_IP);
                udpRcvEP = new IPEndPoint(IPAddress.Any, 0);
            }

            UdpThread = new Thread(new ThreadStart(UdpRecvFunc));
            UdpThread.Start();

            return udpMulticast.Client.IsBound;
        }

        private void UdpRecvFunc()
        {
            m_bUdpThreadAllive = true;

            EndPoint rcvEP = (EndPoint)udpRcvEP;

            while (!m_bUdpTerminate)
            {
                int length = udpMulticast.Client.ReceiveFrom(data, ref rcvEP);

                if (OnUpdate != null && data.Length > 0)
                    OnUpdate(ref data);
            }
            m_bUdpThreadAllive = false;
        }

        public bool GetConnStatus()
        {
            return udpMulticast.Client.IsBound;
        }

        public int SendData(byte[] sendData, int length)
        {
            if (multicastPort == -1)
                return -1;

            return udpMulticast.Send(sendData, length, sendEP);
        }

        public bool Disconnect()
        {
            bool rtn = true;
            int cnt = 0;
            m_bUdpTerminate = true;
            byte[] dummy = new byte[1];
            SendData(dummy, 0);
            while (m_bUdpThreadAllive)
            {
                cnt++;
                if (cnt > 10)
                {
                    rtn = false;
                    break;
                }
                Thread.Sleep(10);
            }
            udpMulticast.Close();
            udpMulticast = null;
            return rtn;
        }

        #region GetRealIpAddress
        private IPAddress GetRealIpAddress()
        {
            IPAddress gateway = FindGetGatewayAddress();

            if (gateway == null)
                return null;

            IPAddress[] pIPAddress = Dns.GetHostAddresses(Dns.GetHostName());

            foreach (IPAddress address in pIPAddress)
                if (IsAddressOfGateway(address, gateway))
                    return address;

            return null;
        }

        private bool IsAddressOfGateway(IPAddress address, IPAddress gateway)
        {
            if (address != null && gateway != null)
                return IsAddressOfGateway(address.GetAddressBytes(), gateway.GetAddressBytes());

            return false;
        }

        private bool IsAddressOfGateway(byte[] address, byte[] gateway)
        {
            if (address != null && gateway != null)
            {
                int gwLen = gateway.Length;
                if (gwLen > 0)
                {
                    if (address.Length == gateway.Length)
                    {
                        --gwLen;
                        int counter = 0;
                        for (int i = 0; i < gwLen; i++)
                        {
                            if (address[i] == gateway[i])
                                ++counter;
                        }
                        return (counter == gwLen);
                    }
                }
            }
            return false;
        }

        private IPAddress FindGetGatewayAddress()
        {
            IPGlobalProperties ipGlobProps = IPGlobalProperties.GetIPGlobalProperties();

            foreach (NetworkInterface ni in NetworkInterface.GetAllNetworkInterfaces())
            {
                IPInterfaceProperties ipInfProps = ni.GetIPProperties();
                foreach (GatewayIPAddressInformation gi in ipInfProps.GatewayAddresses)
                    return gi.Address;
            }
            return null;
        }
        #endregion
    }
}
