﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text.RegularExpressions;
using System.Diagnostics;
using System.Net.NetworkInformation;

namespace cloudtreeReference
{
    public class NetUdpUnicast
    {
        public delegate void PacketUpdateDelegate(ref byte[] data);
        public event PacketUpdateDelegate OnUpdate;

        private UdpClient udpReceiver = null;
        private IPEndPoint udpRcvEP = null;
        private UdpClient udpSender = null;
        private IPEndPoint udpSendEP = null;

        private string rcvIP, sendIP;
        private int rcvPort = -1, sendPort = -1;
        private bool m_bUdpThreadAllive = false;
        private bool m_bUdpTerminate = false;
        private byte[] data;
        private int maxDataLength = -1;

        private Thread UdpThread = null;

        public NetUdpUnicast()
        {
            rcvIP = GetRealIpAddress().ToString();
        }

        public void SetRcvPort(int port)
        {
            rcvPort = port;
        }

        public NetUdpUnicast(string _rcvIP, int _rcvPort, string _sendIP, int _sendPort)
        {
            rcvIP = _rcvIP;
            rcvPort = _rcvPort;
            sendIP = _sendIP;
            sendPort = _sendPort;
        }

        //public String getIP()
        //{
        //    Regex regex = new Regex(@"^(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])\.(\d{1,2}|1\d\d|2[0-4]\d|25[0-5])$");
        //    foreach (System.Net.IPAddress ip in System.Net.Dns.GetHostEntry(System.Net.Dns.GetHostName()).AddressList)
        //    {
        //        if (regex.IsMatch(ip.ToString()))
        //        {
        //            return ip.ToString();
        //        }
        //    }
        //    return null;
        //}

        public void SetDataMaxLength(int length)
        {
            maxDataLength = length;
            data = new byte[length];
        }

        public bool Connect()
        {
            if (udpReceiver != null)
            {
                udpReceiver.Close();
                udpReceiver = null;
            }

            if (udpSender != null)
            {
                udpSender.Close();
                udpSender = null;
            }

            if (rcvPort == -1)
                return false;
            else
            {
                udpReceiver = new UdpClient();
                udpRcvEP = new IPEndPoint(IPAddress.Parse(rcvIP), rcvPort);
                udpReceiver.Client.Bind(udpRcvEP);
            }

            udpSender = new UdpClient();
            if (sendIP.Length > 0 && sendPort != -1)
                udpSendEP = new IPEndPoint(IPAddress.Parse(sendIP), sendPort);

            UdpThread = new Thread(new ThreadStart(UdpRecvFunc));
            UdpThread.Start();

            return udpReceiver.Client.IsBound;
        }

        private void UdpRecvFunc()
        {
            m_bUdpThreadAllive = true;
            while (!m_bUdpTerminate)
            {
                data = udpReceiver.Receive(ref udpRcvEP);

                if (OnUpdate != null && data.Length > 0)
                    OnUpdate(ref data);
            }
            m_bUdpThreadAllive = false;
        }

        public bool GetConnStatus()
        {
            return udpReceiver.Client.IsBound;
        }

        public int SendData(byte[] sendData, int length)
        {
            if (sendPort == -1)
                return -1;

            return udpSender.Send(sendData, length, udpSendEP);
        }

        public int SendData(byte[] sendData, int length, string _sendIP, int _sendPort)
        {
            IPEndPoint _udpSendEP = new IPEndPoint(IPAddress.Parse(_sendIP), _sendPort);
            return udpSender.Send(sendData, length, _udpSendEP);
        }

        public bool Disconnect()
        {
            bool rtn = true;
            int cnt = 0;
            m_bUdpTerminate = true;
            byte[] dummy = new byte[1];
            SendData(dummy, 0, rcvIP, rcvPort);
            while (m_bUdpThreadAllive)
            {
                cnt++;
                if (cnt > 10)
                {
                    rtn = false;
                    break;
                }
                Thread.Sleep(10);
            }
            udpReceiver.Close();
            udpReceiver = null;
            return rtn;
        }

        #region GetRealIpAddress
        private IPAddress GetRealIpAddress()
        {
         IPAddress gateway = FindGetGatewayAddress();

         if (gateway == null)
          return null;

         IPAddress[] pIPAddress = Dns.GetHostAddresses(Dns.GetHostName());

         foreach (IPAddress address in pIPAddress)
          if (IsAddressOfGateway(address, gateway))
           return address;

         return null;
        }

        private bool IsAddressOfGateway(IPAddress address, IPAddress gateway)
        {
         if (address != null && gateway != null)
          return IsAddressOfGateway(address.GetAddressBytes(), gateway.GetAddressBytes());

         return false;
        }

        private bool IsAddressOfGateway(byte[] address, byte[] gateway)
        {
         if (address != null && gateway != null)
         {
          int gwLen = gateway.Length;
          if (gwLen > 0)
          {
           if (address.Length == gateway.Length)
           {
            --gwLen;
            int counter = 0;
            for (int i = 0; i < gwLen; i++)
            {
             if (address[i] == gateway[i])
              ++counter;
            }
            return (counter == gwLen);
           }
          }
         }
         return false;
        }

        private IPAddress FindGetGatewayAddress()
        {
         IPGlobalProperties ipGlobProps = IPGlobalProperties.GetIPGlobalProperties();

         foreach (NetworkInterface ni in NetworkInterface.GetAllNetworkInterfaces())
         {
          IPInterfaceProperties ipInfProps = ni.GetIPProperties();
          foreach (GatewayIPAddressInformation gi in ipInfProps.GatewayAddresses)
           return gi.Address;
         }
         return null;
        }
        #endregion
    }
}
