﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Diagnostics;
using System.Windows.Forms;
using System.Threading;

using System.Net;
using System.Net.Sockets;
using System.Net.NetworkInformation;
using cloudtreeReference;

namespace CloudtreeReference_SampleForm
{
    public partial class Form1 : Form
    {
        CloudtreeReferences m_Reference = null; // 통신 외 수학 및 좌표 등 레퍼런스 클래스
        NetUdpUnicast m_NetUdpUnicast = null; // UDP 유니캐스트 인스턴스
        NetUdpMulticast m_NetUdpMulticast = null; // UDP 멀티캐스트 인스턴스
        NetTcpServer m_NetTcpServer = null; // TCP 서버 인스턴스
        NetTcpClient m_NetTcpClient = null; // TCP 클라이언트 인스턴스

        DateTime startCheck = new DateTime(); // 시간경과 체크 위한 기준시간 체커.

        // 베지어 곡선 그리기 위한 멤버 변수들
        List<PointF> ptList = new List<PointF>(); // 베지어 곡선 그리기 위한 기준점
        List<PointF> BezierResultP = new List<PointF>(); // 베지어 곡선 결과 점 리스트
        List<PointF> CatmullromResultP = new List<PointF>(); // 캣멀롬 곡선 결과 점 리스트
        Graphics g;

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            CenterToScreen();
            g = Graphics.FromHwnd(pictureBox1.Handle);

            m_Reference = new CloudtreeReferences();

            tbTcpServerIP.Text = m_Reference.GetRealIpAddress().ToString();
            tbTcpServerPort.Text = "3001";
            tbUdpRcvIP.Text = m_Reference.GetRealIpAddress().ToString();
            tbUdpSendIP.Text = m_Reference.GetRealIpAddress().ToString();
            tbUdpRcvPort.Text = "3000";
            tbUdpSendPort.Text = "3001";
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            // TCP 서버 소켓 존재 시 서버 닫고 종료.
            if (m_NetTcpServer != null)
            {
                m_NetTcpServer.Disconnect();
                m_NetTcpServer = null;
            }

            // TCP 클라이언트 소켓 존재 시 클라이언트 연결 끊고 종료.
            if (m_NetTcpClient != null)
            {
                m_NetTcpClient.Disconnect();
                m_NetTcpClient = null;
            }

            // UDP unicast 소켓 존재 시 서버 닫고 종료.

            if (m_NetUdpUnicast != null)
            {
                if(m_NetUdpUnicast.Disconnect())
                    m_NetUdpUnicast = null;
            }

            // UDP multicast 소켓 존재 시 서버 닫고 종료.
            if (m_NetUdpMulticast != null)
            {
                m_NetUdpMulticast.Disconnect();
                m_NetUdpMulticast = null;
            }
        }

        private void linkLabel1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            Process.Start("http://cloudtree.tistory.com/");
        }

        #region TCP/IP 서버
        private void btnCreateServer_Click(object sender, EventArgs e)
        {
            // 서버 생성
            m_NetTcpServer = new NetTcpServer(tbTcpServerIP.Text, Convert.ToInt32(tbTcpServerPort.Text));
            m_NetTcpServer.SetDataMaxLength(2048);
            if (!m_NetTcpServer.Connect())
            {
                cbServerState.Checked = false;
                MessageBox.Show("TCP 서버 소켓 생성에 실패했습니다.");
            }
            else
                cbServerState.Checked = true;

            m_NetTcpServer.OnUpdate += OnUpdate_TcpServer;            
        }

        private void OnUpdate_TcpServer(ref byte[] _rcvData)
        {
            string message = Encoding.UTF8.GetString(_rcvData);
            MessageBox.Show(message);
        }

        private void btnCloseServer_Click(object sender, EventArgs e)
        {
            if (m_NetTcpServer != null)
            {
                cbServerState.Checked = false;
                m_NetTcpServer.Disconnect();
                m_NetTcpServer = null;
            }
        }

        private void btnServerSend_Click(object sender, EventArgs e)
        {
            int reLen = -1;
            byte[] StrByte = Encoding.UTF8.GetBytes(tbServerMSG.Text);
            if (m_NetTcpServer != null)
            {
                reLen = m_NetTcpServer.SendData(StrByte, StrByte.Length);
            }
        }
        #endregion


        #region TCP/IP 클라이언트
        private void btnConnect_Click(object sender, EventArgs e)
        {
            // 클라이언트 생성 후 서버 접속
            m_NetTcpClient = new NetTcpClient(tbTcpServerIP.Text, Convert.ToInt32(tbTcpServerPort.Text));
            m_NetTcpClient.SetDataMaxLength(2048);
            if (!m_NetTcpClient.Connect())
            {
                cbConnectionState.Checked = false;
                MessageBox.Show("TCP 서버 소켓 생성에 실패했습니다.");
            }
            else
                cbConnectionState.Checked = true;

            m_NetTcpClient.OnUpdate += OnUpdate_TcpClient;     
            
        }

        private void OnUpdate_TcpClient(ref byte[] _rcvData)
        {
            string message = Encoding.UTF8.GetString(_rcvData);
            MessageBox.Show(message);
        }

        private void btnDisconnect_Click(object sender, EventArgs e)
        {
            // 클라이언트 접속 종료 및 소켓 닫기
            if (m_NetTcpClient != null)
            {
                cbConnectionState.Checked = false;
                m_NetTcpClient.Disconnect();
                m_NetTcpClient = null;
            }
            
        }

        private void btnClientSend_Click(object sender, EventArgs e)
        {
            // 서버로 메세지 전송
            int reLen = -1;
            byte[] StrByte = Encoding.UTF8.GetBytes(tbClientMSG.Text);
            if (m_NetTcpClient != null)
            {
                reLen = m_NetTcpClient.SendData(StrByte, StrByte.Length);
            }
            
        }
        #endregion


        #region UDP 유니캐스트
        private void btnCreateServer_UDP_Click(object sender, EventArgs e)
        {
            m_NetUdpUnicast = new NetUdpUnicast(tbUdpRcvIP.Text, Convert.ToInt32(tbUdpRcvPort.Text), tbUdpSendIP.Text, Convert.ToInt32(tbUdpSendPort.Text));
            m_NetUdpUnicast.SetDataMaxLength(2048);
            if (!m_NetUdpUnicast.Connect())
                MessageBox.Show("UDP 유니캐스트 소켓 생성에 실패했습니다.");

            m_NetUdpUnicast.OnUpdate += OnUpdate_UDPServer;

            /*
            // 인자 없이 m_NetUdpUnicast = new NetUdpUnicast()로 생성 시 수신 포트를 설정해야 하며, 
            // 데이터 송신 시 SendData(byte[] sendData, int length, string _sendIP, int _sendPort)를 사용해 보낼 IP와 Port를 지정해야 함.
            m_NetUdpUnicast = new NetUdpUnicast();
            m_NetUdpUnicast.SetRcvPort(Convert.ToInt32(tbUdpRcvPort.Text));
            m_Reference.SetDataMaxLength(2048);
            if (!m_NetUdpUnicast.Connect())
                MessageBox.Show("UDP 유니캐스트 소켓 생성에 실패했습니다.");
            m_NetUdpUnicast.OnUpdate += OnUpdate_UDPServer;
              ...
            // 데이터 전송 예제 : m_NetUdpUnicast.SendData(Data, idx, "192.168.100.99", 7001);
            */
        }

        private void OnUpdate_UDPServer(ref byte[] _rcvData)
        {
            string message = Encoding.UTF8.GetString(_rcvData);
            MessageBox.Show(message);
        }

        private void btnCloseServer_UDP_Click(object sender, EventArgs e)
        {
            if (m_NetUdpUnicast != null)
            {
                m_NetUdpUnicast.Disconnect();
                m_NetUdpUnicast = null;
            }
        }

        private void btnServerSend_UDP_Click(object sender, EventArgs e)
        {
            int reLen = -1;
            byte[] StrByte = Encoding.UTF8.GetBytes(tbServerMSG_UDP.Text);
            if (m_NetUdpUnicast != null)
            {
                reLen = m_NetUdpUnicast.SendData(StrByte, StrByte.Length);
            }
        }
        #endregion


        #region UDP 멀티캐스트
        private void btnCreateServer_UDP_Multicast_Click(object sender, EventArgs e)
        {
            m_NetUdpMulticast = new NetUdpMulticast(tbUdpMulticastIP.Text, Convert.ToInt32(tbUdpMulticastPort.Text));
            m_NetUdpMulticast.SetDataMaxLength(2048);
            if (!m_NetUdpMulticast.Connect())
                MessageBox.Show("UDP 멀티캐스트 소켓 생성에 실패했습니다.");

            m_NetUdpMulticast.OnUpdate += OnUpdate_UDPServer;
        }

        private void btnCloseServer_UDP_Multicast_Click(object sender, EventArgs e)
        {
            if (m_NetUdpMulticast != null)
            {
                m_NetUdpMulticast.Disconnect();
                m_NetUdpMulticast = null;
            }
        }

        private void btnServerSend_UDP_Multicast_Click(object sender, EventArgs e)
        {
            int reLen = -1;
            byte[] StrByte = Encoding.UTF8.GetBytes("Sender[" + m_iSender.ToString() + "] / Receiver[" + m_iReceiver.ToString() + "] : " + (tbServerMSG_UDP.Text));
            if (m_NetUdpMulticast != null)
            {
                reLen = m_NetUdpMulticast.SendData(StrByte, StrByte.Length);
            }
        }
        #endregion


        #region 수학함수
        private void btnEuclideanDist2D_Click(object sender, EventArgs e)
        {
            double distance = m_Reference.EuclideanDistance(Convert.ToDouble(tbEuclidean2D_x1.Text), Convert.ToDouble(tbEuclidean2D_y1.Text), Convert.ToDouble(tbEuclidean2D_x2.Text), Convert.ToDouble(tbEuclidean2D_y2.Text));
            tbEuclidean2D_Dist.Text = distance.ToString();
        }

        private void btnEuclideanDist3D_Click(object sender, EventArgs e)
        {
            double distance = m_Reference.EuclideanDistance(Convert.ToDouble(tbEuclidean3D_x1.Text), Convert.ToDouble(tbEuclidean3D_y1.Text), Convert.ToDouble(tbEuclidean3D_z1.Text), Convert.ToDouble(tbEuclidean3D_x2.Text), Convert.ToDouble(tbEuclidean3D_y2.Text), Convert.ToDouble(tbEuclidean3D_z2.Text));
            tbEuclidean3D_Dist.Text = distance.ToString();
        }

        private void btnGaussianRandom_Click(object sender, EventArgs e)
        {
            double randomValue = m_Reference.Gaussian(Convert.ToDouble(tbGaussianRandom_mean.Text), Convert.ToDouble(tbGaussianRandom_stdev.Text));
            tbGaussianRandom_result.Text = randomValue.ToString();
        }
        #endregion


        #region 좌표계
        private void btnLatLonBearing_Click(object sender, EventArgs e)
        {
            double bearing = m_Reference.bearingP1toP2(Convert.ToDouble(tbLat1.Text), Convert.ToDouble(tbLon1.Text), Convert.ToDouble(tbLat2.Text), Convert.ToDouble(tbLon2.Text));
            tbLatLonBearing.Text = bearing.ToString();
        }

        private void btnLatLonDistance_Click(object sender, EventArgs e)
        {
            double distance = m_Reference.LatLonDistance(Convert.ToDouble(tbLat1.Text), Convert.ToDouble(tbLon1.Text), Convert.ToDouble(tbLat2.Text), Convert.ToDouble(tbLon2.Text));
            tbLatLonDistance.Text = distance.ToString();
        }

        private void btnECEF2LLA_Click(object sender, EventArgs e)
        {
            double lat = 0.0;
            double lon = 0.0;
            double alt = 0.0;

            m_Reference.ECEF2LatLonAlt(Convert.ToDouble(tbTransX.Text), Convert.ToDouble(tbTransY.Text), Convert.ToDouble(tbTransZ.Text), ref lat, ref lon, ref alt);
            tbTransLat.Text = lat.ToString();
            tbTransLon.Text = lon.ToString();
            tbTransAlt.Text = alt.ToString();
        }
        #endregion


        #region 시간 경과
        private void btnStartCheck_Click(object sender, EventArgs e)
        {
            startCheck = DateTime.Now;
        }

        private void btnGetDuration_Click(object sender, EventArgs e)
        {
            double duration = m_Reference.GetDuration(startCheck);
            tbDurationTimeSec.Text = duration.ToString();
        }
        #endregion


        #region 공유메모리
        private void btnCreateSharedMem_Click(object sender, EventArgs e)
        {
            string key = tbSharedMemKey1.Text;
            byte[] data = Encoding.UTF8.GetBytes(tbSharedMemData.Text);
            m_Reference.CreateSharedMemory(ref key);
        }

        private void btnDestroySharedMem_Click(object sender, EventArgs e)
        {
            m_Reference.CloseSharedMem();
        }

        private void btnWriteSharedMemData_Click(object sender, EventArgs e)
        {
            string key = tbSharedMemKey1.Text;
            byte[] data = Encoding.UTF8.GetBytes(tbSharedMemData.Text);
            m_Reference.WriteSharedMemory(ref key, ref data);
        }

        private void btnRefreshSharedMemData_Click(object sender, EventArgs e)
        {
            string key = tbSharedMemKey2.Text;
            byte[] data = new byte[2048];
            m_Reference.ReadSharedMemory(ref key, ref data);
            tbSharedMemDataRead.Text = Encoding.Default.GetString(data);
        }
        #endregion


        #region 보간곡선
        private void pictureBox1_MouseClick(object sender, MouseEventArgs e)
        {
            PointF newP = new PointF();
            newP.X = e.X;
            newP.Y = e.Y;
            ptList.Add(newP);

            Pen px = new Pen(Brushes.Red);
            g.DrawRectangle(px, new Rectangle(e.X, e.Y, 1, 1));
        }

        private void btnDrawBezierCurve_Click(object sender, EventArgs e)
        {
            if (ptList.Count < 2)
                MessageBox.Show("2개 이상의 곡선 기준점을 찍어주세요.");
            else
            {
                int NumOfCurvePoints = Convert.ToInt32(tbNumOfCurvePoints.Text);
                List<PointF> pts = new List<PointF>(ptList);

                m_Reference.getBezier2DPointList(ref pts, NumOfCurvePoints, ref BezierResultP);

                PointF[] splinePoints = new PointF[BezierResultP.Count];
                for (int i = 0; i < BezierResultP.Count; i++)
                    splinePoints[i] = BezierResultP[i];

                for (int i = 0; i < NumOfCurvePoints; i++)
                {
                    g.DrawLines(Pens.Magenta, splinePoints);
                    g.Flush();
                    Application.DoEvents();
                }
            }
        }

        private void btnClear_Click(object sender, EventArgs e)
        {
            ptList.Clear();
            BezierResultP.Clear();
            CatmullromResultP.Clear();
            g.Clear(Color.White);
        }

        private void btnDrawCatmullRomCurve_Click(object sender, EventArgs e)
        {
            if (ptList.Count < 2)
                MessageBox.Show("2개 이상의 곡선 기준점을 찍어주세요.");
            else
            {
                List<PointF> pts = new List<PointF>(ptList);
                m_Reference.getCatmullRom2DPointList(ref pts, ref CatmullromResultP);
                PointF[] splinePoints = new PointF[CatmullromResultP.Count];
                for (int i = 0; i < CatmullromResultP.Count; i++)
                    splinePoints[i] = CatmullromResultP[i];

                for (int i = 0; i < CatmullromResultP.Count; i++)
                {
                    g.DrawLines(Pens.Green, splinePoints);
                    g.Flush();
                    Application.DoEvents();
                }
            }
        }
        #endregion


        private int m_iSender = 0;
        private int m_iReceiver = 0;

        public interface INetFuncs
        {
            int get_Buffer(ref byte[] bufer);
        }

        public class MsgSample_0001 : INetFuncs
        {
            public int sender = 0;
            public int receiver = 0;
            public byte[] byteMsg = new byte[1024];
            public MsgSample_0001() { }

            public int get_Buffer(ref byte[] buffer)
            {
                int idx = 0;
                byte[] data = new byte[2048];

                Buffer.BlockCopy(BitConverter.GetBytes(sender), 0, data, idx, 4); idx += 4;
                Buffer.BlockCopy(BitConverter.GetBytes(receiver), 0, data, idx, 4); idx += 4;
                for(int i=0;i<1024;i++)
                {
                    try
                    {
                        Buffer.BlockCopy(BitConverter.GetBytes(byteMsg[i]), 0, data, idx, 1); idx += 1;
                    }
                    catch (Exception ex) { break; }
                }
                Buffer.BlockCopy(data, 0, buffer, 0, idx);

                return idx;
            }

            public void Parse(ref byte[] buffer, ref int idx)
            {
                sender = BitConverter.ToInt32(buffer, idx); idx += 4;
                receiver = BitConverter.ToInt32(buffer, idx); idx += 4;
                for (int i = 0; i < 1024; i++)
                {
                    byteMsg[i] = (byte)BitConverter.ToChar(buffer, idx); idx += 1;
                }
            }

        }

        public struct _Header
        {
            public UInt32 OPCODE;
            public UInt32 PacketSize;
            public int sender;
            public int receiver;
        }

        public struct _msg4202
        {
            public int TargetID;
        }
        

        private void rb_ID1_CheckedChanged(object sender, EventArgs e)
        {
            m_iSender = 1;
            if (rb_ID1.Checked)
            {
                rb_SendTo1.Enabled = false;
                rb_SendTo2.Enabled = true;
                rb_SendTo3.Enabled = true;
            }
        }

        private void rb_ID2_CheckedChanged(object sender, EventArgs e)
        {
            m_iSender = 2;
            if (rb_ID2.Checked)
            {
                rb_SendTo1.Enabled = true;
                rb_SendTo2.Enabled = false;
                rb_SendTo3.Enabled = true;
            }
        }

        private void rb_ID3_CheckedChanged(object sender, EventArgs e)
        {
            m_iSender = 3;
            if (rb_ID3.Checked)
            {
                rb_SendTo1.Enabled = true;
                rb_SendTo2.Enabled = true;
                rb_SendTo3.Enabled = false;
            }
        }

        private void rb_SendTo1_CheckedChanged(object sender, EventArgs e)
        {
            m_iReceiver = 1;
        }

        private void rb_SendTo2_CheckedChanged(object sender, EventArgs e)
        {
            m_iReceiver = 2;
        }

        private void rb_SendTo3_CheckedChanged(object sender, EventArgs e)
        {
            m_iReceiver = 3;
        }

        private void rb_SendToAll_CheckedChanged(object sender, EventArgs e)
        {
            m_iReceiver = 4;
        }

        

        

    }
}
